from pydantic import BaseModel

from src.domain._shared import inherit_field
from src.domain.todo import TodoEntity


class CreateTodoBodyDTO(BaseModel):
    title: str = inherit_field(TodoEntity, "title")
    description: str | None = inherit_field(TodoEntity, "description")



from abc import ABC, abstractmethod
from collections.abc import Awaitable, Callable
from dataclasses import dataclass
from typing import Any


class Database(ABC):
    @dataclass(frozen=True)
    class Result:
        rows_affected: int
        last_insert_id: str | int | None = None

    class Transaction(ABC):
        @abstractmethod
        async def query[T](
            self,
            sql: str,
            params: list[Any] | None = None,
        ) -> list[T]:
            raise NotImplementedError

        @abstractmethod
        async def exec(
            self,
            sql: str,
            params: list[Any] | None = None,
        ) -> Database.Result:
            raise NotImplementedError

    @abstractmethod
    async def ping(self) -> None:
        raise NotImplementedError

    @abstractmethod
    async def connect(self) -> None:
        raise NotImplementedError

    @abstractmethod
    async def close(self) -> None:
        raise NotImplementedError

    @abstractmethod
    async def query[T](
        self,
        sql: str,
        params: list[Any] | None = None,
    ) -> list[T]:
        raise NotImplementedError

    @abstractmethod
    async def exec(
        self,
        sql: str,
        params: list[Any] | None = None,
    ) -> Database.Result:
        raise NotImplementedError

    @abstractmethod
    async def transaction[T](
        self,
        handler: Callable[[Database.Transaction], Awaitable[T]],
    ) -> T:
        raise NotImplementedError



from typing import Any


class Inject:
    """Marker used in ``Annotated`` type hints to specify a custom injection token.

    When a parameter is annotated with ``Annotated[SomeType, Inject(token)]``,
    the container resolves ``token`` instead of ``SomeType``.

    Example::

        async def handler(
            svc: Annotated[MyService, Inject("my_service_token")]
        ) -> None: ...
    """

    def __init__(self, token: Any) -> None:
        """
        Args:
            token: The token (class, string, or any hashable) that the container
                   should resolve for the annotated parameter.
        """
        self.token = token



from dijay import module

from .create_todo import CreateTodoUsecase
from .delete_todo import DeleteTodoUsecase
from .get_todo_by_id import GetTodoByIdUsecase
from .list_todo import ListTodoUsecase
from .update_todo import UpdateTodoUsecase

providers = [
    CreateTodoUsecase,
    DeleteTodoUsecase,
    GetTodoByIdUsecase,
    ListTodoUsecase,
    UpdateTodoUsecase,
]


@module(providers=providers, exports=providers)
class TodoModule:
    pass



from fastapi import APIRouter

from src.application.todo import (
    CreateTodoData,
    CreateTodoUsecase,
    DeleteTodoData,
    DeleteTodoUsecase,
    GetTodoByIdData,
    GetTodoByIdUsecase,
    ListTodoResult,
    ListTodoUsecase,
    UpdateTodoData,
    UpdateTodoUsecase,
)
from src.domain.todo import TodoEntity, TodoNotFoundError
from src.presentation.http._shared import inject, map_domain_error
from src.presentation.http.todo.dtos import CreateTodoBodyDTO, UpdateTodoBodyDTO

router = APIRouter(prefix="/todos", tags=["todos"])


@router.post("/", status_code=201)
async def create_todo(
    body: CreateTodoBodyDTO,
    usecase: inject(CreateTodoUsecase),
) -> TodoEntity:
    data = CreateTodoData(
        title=body.title,
        description=body.description,
    )
    return await usecase.execute(data)


@router.get("/")
async def list_todos(
    usecase: inject(ListTodoUsecase),
) -> ListTodoResult:
    return await usecase.execute()


@router.get("/{id}")
@map_domain_error((TodoNotFoundError, 404))
async def get_todo(
    id: str,
    usecase: inject(GetTodoByIdUsecase),
) -> TodoEntity:
    data = GetTodoByIdData(
        id=id,
    )
    return await usecase.execute(data)


@router.put("/{id}")
@map_domain_error((TodoNotFoundError, 404))
async def update_todo(
    id: str,
    body: UpdateTodoBodyDTO,
    usecase: inject(UpdateTodoUsecase),
) -> TodoEntity:
    data = UpdateTodoData(
        id=id,
        changes=UpdateTodoData.Changes(
            title=body.title,
            description=body.description,
            done=body.done,
        ),
    )
    return await usecase.execute(data)


@router.delete("/{id}", status_code=204)
@map_domain_error((TodoNotFoundError, 404))
async def delete_todo(
    id: str,
    usecase: inject(DeleteTodoUsecase),
):
    data = DeleteTodoData(
        id=id,
    )
    await usecase.execute(data)
